/// @author Stanisaw Aleksander Janikowski (stasiek.janikowski@gmail.com)
/* 
 * Program stworzony w ramach pracy domowej przedmiotu ZPR w celu 
 * demonstracji i przyblizenia algorytmu transform dostepnego w bibliotece
 * standardowej (stl)
 */

// Szablon funkcji transform dostpny jest w pliku nagwkowym:
#include<algorithm>

// Pliki nagwkowe, wykorzystywane pomocniczo w przykadach:
#include<iostream>
#include<vector>
#include<map>
#include<complex>
#include<cmath>

using namespace std;

/*
 * Algorytm transform suy do dokonywania tego samego przeksztacenia na 
 * wielu elementach zawartych w kontenerze, a take do wykonania operacji
 * dwuargumentowych na odpowiadajcych sobie elementach z dwch kontenerw
 * Wynik przeksztacenia moe by zapisywany w miejscu, nadpisujc dane w
 * jednym z kontenerw wejciowych, lub do innego kontenera.
 */

/*
 * Wersja z operatorem jednoargumentowym:
 *	template < class InputIterator, class OutputIterator, class UnaryOperator >
 *		OutputIterator transform ( InputIterator first1, InputIterator last1,
 *					   OutputIterator result, UnaryOperator op );
 */

// Przykad 1 (wystawianie ocen)
// funkcja pomocnicza:
float wystaw_ocene(int punkty){
	punkty-=51;
	if (punkty < 0) return 2.0;
	return 3.0 + 0.5 * (punkty/10);		
}

// Nic nie stoi na przeszkodzie, aby transformacj robi w miejscu (wyniki nadpisuj argumenty)
// Przykad 2 (przeliczanie jednostek)
// funkcja pomocnicza:
float Cels_to_Fahr(float temp){
	return 32.0 + 9.0/5.0 * temp;
}

/*
 * Wersja z operatorem dwuargumentowym:
 * 	template < class InputIterator1, class InputIterator2,
 * 		   class OutputIterator, class BinaryOperator >
 * 		OutputIterator transform ( InputIterator1 first1, InputIterator1 last1, 
 * 					   InputIterator2 first2, OutputIterator result, 
 * 					   BinaryOperator binary_op );
 */

// Przykad 3 (widmo zespolone)
// Widmo amplitudowe i fazowe zapisane w oddzielnych kontenerach integrowane do widma zespolonego
typedef complex<double> dcomplex;

dcomplex licz_widmo(double ampl, double faza){
	return dcomplex(ampl * cos(faza), ampl * sin(faza) );
}

/*
 * Transform ma due moliwoci, lecz jego stosowalno jest ograniczona wymaganiem standardw, aby
 * dostarczona funkcja bya pozbawiona efektw ubocznych. Np. nie moemy uy transform do edycji
 * mapy poniewa niedozwolone jest uycie operatora przypisania, a zmiana wartoci przez funkcj
 * naruszaaby standard. Podobnie nie moemy uywa transform aby usuwa lub przenosi elementy.
 * Kolejnym ograniczeniem jest brak gwarancji kolejnoci wykonywania operacji na elementach, wic
 * uycie transform z uyciem funkcji zalenej od kolejnoci woania jest bdem. W wymienionych
 * przypadkach naley stosowa funkcj for_each lub stworzy wasn, nie obarczon tymi warunkami
 * wersj algorytmu transform.
 *	
 * UWAGA: W wielu kompilatorach kod ignorujcy wspomniane wyej zaoenia zadziaa, jednak nie ma
 * 	  adnej gwarancji jego przenonoci, rwnie pomidzy wersjami tego samego kompilatora.
 * 	  Wspomniane ograniczenia s spowodowane deniem do umoliwienia kompilatorom optymalizacji.
 * 	  W szczeglnoci operacje na poszczeglnych elementach zakresu funkcji transform mog by
 * 	  wykonywane w osobnych wtkach.
 */


int main(){

	// Przykad 1 (wystawianie ocen):
	cout.setf(ios::fixed,ios::floatfield);
	cout.precision(1);
	vector<int> punkty;
	vector<float> oceny;
	vector<int>::iterator it_p;
	vector<float>::iterator it_o;
	
	// Ustawienie i wypisanie przykadowych punktw uzyskanych przez studentw
	for (int i=1; i<10; i++) punkty.push_back (i*51%81 + 20);
	cout << "Przyklad 1:\nZdobyte punkty:";
	for (it_p=punkty.begin(); it_p!=punkty.end(); ++it_p)
		cout << "\t" << *it_p;
	cout << endl;

	//Rozszerzenie bufora docelowego (oceny), aby zmiesci wartosci:
	oceny.resize(punkty.size());
	// Uycie transform do przeliczenia punktw na oceny:
	transform(punkty.begin(), punkty.end(), oceny.begin(), wystaw_ocene);

	//Wypisanie ocen:
	cout << "Uzyskane Oceny:";
	for (it_o=oceny.begin(); it_o!=oceny.end(); ++it_o)
		cout << "\t" << *it_o;
	cout << endl;

/*****************************************************************************************************/

	// Przykad 2 (przeliczanie jednostek):
	cout.precision(2);
	vector<float> temperatura;
	vector<float>::iterator it;
	
	// Ustawienie i wypisanie przykadowych wartosci temperatury w stopniach Celsiusza
	temperatura.push_back(36.6);
	for (int i=1; i<9; i++) temperatura.push_back (i*249%611 /15.0);
	cout << "Przyklad 2:\nTemperatury w st.C:";
	for (it=temperatura.begin(); it!=temperatura.end(); ++it)
		cout << "\t" << *it;
	cout << endl;
	
	// Uycie transform do przeliczenia temperatur
	transform(temperatura.begin(), temperatura.end(), temperatura.begin(), Cels_to_Fahr);
	
	// Wypisanie temperatur w stopniach Fahrenheita
	cout << "Temperatury w st.F:";
	for (it=temperatura.begin(); it!=temperatura.end(); ++it)
		cout << "\t" << *it;
	cout << endl;

/*****************************************************************************************************/

	// Przykad 3 (widmo zespolone):
	cout.precision(3);
	vector<double> widmo_ampl;
	vector<double> widmo_faz;
	vector<dcomplex> widmo;
	vector<double>::iterator it1;
	vector<dcomplex>::iterator it2;


	// Ustawienie i wypisanie przykadowych wartosci widma:
	for (int i=0; i<9; i++){
		widmo_ampl.push_back(i%5?5.0/(i*i):0);
		widmo_faz.push_back(M_PI * sin(2 * i%5));
	}
	cout << "Przyklad 3:\nWidmo amplitudowe:";
	for (it1=widmo_ampl.begin(); it1!=widmo_ampl.end(); ++it1)
		cout << "\t" << *it1;
	cout << endl << "Widmo fazowe:	";
	for (it1=widmo_faz.begin(); it1!=widmo_faz.end(); ++it1)
		cout << "\t" << *it1;

	//Rozszerzenie bufora docelowego (widmo), aby zmiesci wartosci:
	widmo.resize(widmo_ampl.size());
	// Uycie transform do obliczenia widma zespolonego
	transform(widmo_ampl.begin(), widmo_ampl.end(), widmo_faz.begin(), widmo.begin(), licz_widmo);
	
	// Wypisanie widma zespolonego
	cout << endl << "Widmo zespolone:";
	for (it2=widmo.begin(); it2!=widmo.end(); ++it2)
		cout << "\t" << *it2;
	cout << endl;

	return 0;
}
